/*
 *  driver for the VIP1024 VME frame grabber video card
 *	forsyth@caldo.demon.co.uk
 */

#include	"u.h"
#include	"../port/lib.h"
#include	"mem.h"
#include	"dat.h"
#include	"fns.h"
#include	"io.h"
#include	"../port/error.h"

#include	"devtab.h"

#define	VIDBASE	0x01000000	/* physical address of VIP on VME32 */

/* offsets from VIDBASE */
#define	VIDBUF	0x0		/* frame store */
#define	VIDREG	0x00100000	/* base address of registers */

/* offset from VIDREG */
#define	VIDCTL0	0x2800

#define	FRAMESIZE (512*512)
#define	FRAMEBUF(i) (framebuffer + (((i)&3)*FRAMESIZE))

enum {
	Inlut = 0,
	Rlut = 1,
	Glut = 2,
	Blut = 3,
	LUTSIZE = 0x800,
	LUTMAPSIZE = 256,
};

typedef struct Video Video;
struct Video {
	uchar	ctl0;	/* 2800 */
	uchar	pad0[0x100-1];
	uchar	ctl1;	/* 2900 */
	uchar	pad1[0x100-1];
	uchar	mask;	/* 2a00 */
	uchar	pad2[0x100-1];
	uchar	status;	/* 2b00 */
	uchar	pad3[0x100-1];
	uchar	snap;	/* 2c00 */
	uchar	pad4[0x100*4-1];
	uchar	crtca;	/* 3000 */
	uchar	pad5[0x100*4-1];
	uchar	crtcd;	/* 3400 */
	uchar	pad6[0x100*4-1];
	uchar	gain;	/* 3800 */
	uchar	pad7[0x100*4-1];
	uchar	offset;	/* 3c00 */
};

enum {
				/* Control Register 0 */
	LutSel = 7,		/* LUT table map select (mask) */
	Quadbuf = 1<<3,	/* =1, four 512x512 frames; =0, one 1Mbyte frame */
	Quadno = 3<<4,	/* active quadrant */
	RASDisable = 1<<7,	/* prevent memory refresh */

				/* Control Register 1 */
	INPsel = 3,		/* input selection (mask) */
	SelALU = 1<<2,		/* Select ALU (select video source) */
	KeyEnable = 1<<3,	/* select video source */
	VideoDisable = 1<<4,	/* video output disable */
	Continuous = 1<<5,	/* =0, snapshot mode */
	EurStandard = 1<<6,	/* =1, 50Hz, 625 lines; =0, 60Hz, 525 lines */
	ExtSync = 1<<7,		/* =1, select external sync */


				/* Status Register */
	VMESlave = 1<<0,	/* =0, VME master */
	Grabbing = 1<<1,	/* still grabbing a frame or frames */
	StatusStrap = 1<<2,	/* =0, jumper 43-46 is IN */
	OddField = 1<<3,	/* =1, display frame in odd field */
};

#define	CWR(x,v)	(vid->x = softvid.x = (v))
#define	CRD(x)	softvid.x

enum{
	Qdir,
	Qcrtc,
	Qvid0,
	Qvid1,
	Qvid2,
	Qvid3,
	Qsnap,
	Qrgb,
	Qreg,
	Qctl,
};
Dirtab vidtab[]={
	"vidcrtc",	{Qcrtc},	16,	0600,
	"vid0",		{Qvid0},	FRAMESIZE,	0600,
	"vid1",		{Qvid1},	FRAMESIZE,	0600,
	"vid2",		{Qvid2},	FRAMESIZE,	0600,
	"vid3",		{Qvid3},	FRAMESIZE,	0600,
	"vidsnap",	{Qsnap},	FRAMESIZE,	0400,
	"vidrgb",	{Qrgb},		LUTMAPSIZE*4,	0600,
	"vidreg",	{Qreg},		sizeof(Video),	0600,
	"vidctl",	{Qctl},		0,	0600,
};
#define Nvidtab (sizeof(vidtab)/sizeof(Dirtab))

enum {
	USvision,
	Eurovision,
	Nformat,

	Ncrtcreg = 16
};

static	uchar	crtcparam[Nformat][Ncrtcreg] = {
	{0x51, 0x40, 0x42, 0x4e, 0x1f, 0x06, 0x1e, 0x1e, 0x07, 0x0f,
	 0, 0, 0, 0, 0, 0 },
	{0x52, 0x40, 0x43, 0x8e, 0x26, 0x00, 0x20, 0x23, 0x07, 0x0f,
	 0, 0, 0, 0, 0, 0 },
};

static	uchar	*vidbase;
static	Video	*vidregs;
static	uchar	*framebuffer;
static	int	curquad;
static	Video	softvid;	/* driver's copy, for write-only registers */
static	uchar	softcrtc[Ncrtcreg];

static	void	initcrtc(int);
static	void	luts(int, int, int, int);
static	void	snap(int);
static	void	setsnap(int);
static	void	setcontin(int);
static	void	vmecpy(void*, void*, long);
static	void	resetstate(void);

void
vidreset(void)
{
	print("VIP-1024 reset\n");
	if(vidbase == 0) {
		vidbase = (uchar*)VA(kmapvme(VIDBASE));
		if(vidbase == 0)
			return;
	}
	vidregs = (Video*)(vidbase+VIDREG+VIDCTL0);
	framebuffer = (uchar*)(vidbase+VIDBUF);
	resetstate();
}

static void
resetstate(void)
{
	Video *vid = vidregs;

	CWR(ctl0, Quadbuf);
	curquad = 0;
	CWR(ctl1, ExtSync|EurStandard|VideoDisable|SelALU);
	initcrtc(Eurovision);
	luts(0, Inlut, 0xff, 0xff);
	luts(0, Rlut, 1, 0);
	luts(0, Glut, 1, 0);
	luts(0, Blut, 1, 0);
	CWR(ctl1, softvid.ctl1 & ~VideoDisable);
	CWR(gain, 0xc0);
	CWR(offset, 0x40);
	CWR(mask,0);
}

void
vidinit(void)
{
}

Chan *
vidattach(char *spec)
{
	if(vidbase == 0)
		error(Enodev);
	return devattach('v', spec);
}

Chan *
vidclone(Chan *c, Chan *nc)
{
	return devclone(c, nc);
}

int
vidwalk(Chan *c, char *name)
{
	return devwalk(c, name, vidtab, (long)Nvidtab, devgen);
}

void
vidstat(Chan *c, char *db)
{
	devstat(c, db, vidtab, (long)Nvidtab, devgen);
}

Chan *
vidopen(Chan *c, int omode)
{
	if(c->qid.path == CHDIR || c->qid.path == Qsnap){
		if(omode != OREAD)
			error(Eperm);
	}
	c->mode = openmode(omode);
	c->flag |= COPEN;
	c->offset = 0;
	return c;
}

void
vidcreate(Chan *c, char *name, int omode, ulong perm)
{
	USED(c, name, omode, perm);
	error(Eperm);
}

void
vidremove(Chan *c)
{
	USED(c);
	error(Eperm);
}

void
vidwstat(Chan *c, char *dp)
{
	USED(c, dp);
	error(Eperm);
}

void
vidclose(Chan *c)
{
	USED(c);
}

long
vidread(Chan *c, void *a, long n, ulong offset)
{
	Video *vid = vidregs;
	int m;

	switch((int)(c->qid.path&~CHDIR)){
	case Qdir:
		return devdirread(c, a, n, vidtab, Nvidtab, devgen);
	case Qcrtc:
		if(offset >= Ncrtcreg || n < 0)
			return 0;
		if(offset+n > Ncrtcreg) {
			n = Ncrtcreg-offset;
			if(n < 0)
				n = 0;
		}
		for(m = 0; m < n; m++, offset++)
			((uchar*)a)[m] = softcrtc[offset];
		break;
	case Qvid0:
	case Qvid1:
	case Qvid2:
	case Qvid3:
		if(n < 0)
			return 0;
		if(offset+n > FRAMESIZE) {
			n = FRAMESIZE-offset;
			if(n < 0)
				n = 0;
		}
		vmecpy(a, FRAMEBUF(c->qid.path-Qvid0)+offset, n);
		break;
	case Qsnap:
		if(n < 0)
			return 0;
		if(offset == 0) {
			snap(curquad);
			if((softvid.ctl1 & Continuous) == 0) {
				while(vid->status & Grabbing)
					;	/* should tsleep? how long does this take? */
			}
		}
		if(offset+n > FRAMESIZE) {
			n = FRAMESIZE-offset;
			if(n < 0)
				n = 0;
		}
		vmecpy(a, FRAMEBUF(curquad)+offset, n);
		break;
	case Qreg:
		if(offset+n > sizeof(Video))
			n = sizeof(Video)-offset;
		if(n < 0)
			n = 0;
		softvid.status = vid->status;
		memmove(a, (uchar*)&softvid+offset, n);
		break;
	case Qrgb:
	case Qctl:
	default:
		n=0;
		break;
	}
	return n;
}

long
vidwrite(Chan *c, char *a, long n, ulong offset)
{
	Video *vid = vidregs;
	char buf[32];
	char *field[5];
	int m;
	uchar *p, *q;

	switch((int)(c->qid.path&~CHDIR)){
	case Qcrtc:
		if(offset >= Ncrtcreg || n < 0)
			return 0;
		if(offset+n > Ncrtcreg) {
			n = Ncrtcreg-offset;
			if(n < 0)
				n = 0;
		}
		for(m = 0; m < n; m++, offset++) {
			vid->crtca = offset;
			vid->crtcd = softcrtc[offset] = ((uchar*)a)[m];
		}
		break;
	case Qvid0:
	case Qvid1:
	case Qvid2:
	case Qvid3:
		if(n < 0)
			return 0;
		if(offset+n > FRAMESIZE) {
			n = FRAMESIZE-offset;
			if(n < 0)
				n = 0;
		}
		vmecpy(FRAMEBUF(c->qid.path-Qvid0)+offset, a, n);
		break;
	case Qctl:
		if(n > sizeof(buf)-1)
			n = sizeof(buf)-1;
		memmove(buf, a, n);
		buf[n] = 0;
		m = getfields(buf, field, 5, " ");
		if(m < 1)
			error(Ebadarg);
		if(strcmp(field[0], "mode") == 0) {
			void (*f)(int);

			if(m < 2)
				error(Ebadarg);
			if(strcmp(field[1], "snap") == 0)
				f = setsnap;
			else if(strncmp(field[1], "contin", 6) == 0)
				f = setcontin;
			else
				error(Ebadarg);
			if(m < 3)
				(*f)(curquad);
			else
				(*f)(atoi(field[2]));
		} else if(strcmp(field[0], "snap") == 0) {
			if(m < 2)
				snap(curquad);
			else
				snap(atoi(field[1]));
		} else if(strcmp(field[0], "format") == 0) {
			if(m < 2)
				error(Ebadarg);
			if(strcmp(field[1], "us") == 0)
				initcrtc(USvision);
			else if(strcmp(field[1], "eur") == 0)
				initcrtc(Eurovision);
			else
				error(Ebadarg);
		} else if(strcmp(field[0], "video") == 0) {
			if(m < 1 || strcmp(field[1], "on") == 0)
				CWR(ctl1, softvid.ctl1 & ~VideoDisable);
			else if(strcmp(field[1], "off") == 0)
				CWR(ctl1, softvid.ctl1 | VideoDisable);
		} else if(strcmp(field[0], "gain") == 0) {
			if(m > 1)
				vid->gain = strtoul(field[1], 0, 0);
		} else if(strcmp(field[0], "offset") == 0) {
			if(m > 1)
				vid->offset = strtoul(field[1], 0, 0);
		} else if(strcmp(field[0], "reset") == 0) {
			resetstate();
		} else if(*field[0] == 'f') {
			switch(field[0][1]) {
			case '0':
			case '1':
			case '2':
			case '3':
				CWR(ctl1, (softvid.ctl1 & ~(SelALU|KeyEnable)) | (field[0][1]-'0')<<2);
				break;
			}
		} else if(*field[0] == 'i') {
			switch(field[0][1]) {
			case '0':
			case '1':
			case '2':
			case '3':
				CWR(ctl1, (softvid.ctl1 & ~INPsel) | (field[0][1]-'0'));
				break;
			}
		} else if(*field[0] == 'l') {
			switch(field[0][1]) {
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
				CWR(ctl0, (softvid.ctl0 & ~LutSel) | (field[0][1]-'0'));
				break;
			}
		} else
			error(Ebadarg);
		break;
	case Qreg:
		if(offset+n > sizeof(Video))
			n = offset-sizeof(Video);
		if(n > 0) {
			/* BUG: should copy just the register values */
			memmove((uchar*)&softvid+offset, a, n);
			vid->ctl0 = softvid.ctl0;
			vid->ctl1 = softvid.ctl1;
			vid->mask = softvid.mask;
			softvid.status = vid->status;	/* status is read only */
			vid->gain = softvid.gain;
			vid->offset = softvid.offset;
		}
		break;
	case Qrgb:
		if(offset+n > 2048)
			n = 2048-offset;
		if(n < 0)
			n = 0;
		p = vidbase+VIDREG;
		q = (uchar*)a;
		while(--n >= 0)
			*p++ = *q++;	/* device only supports byte access */
		break;
	default:
		error(Ebadusefd);
	}
	return n;
}

static void
initcrtc(int mode)
{
	int i;
	Video *vid = vidregs;

	if(mode < 0 || mode >= Nformat)
		error(Ebadarg);
	for(i=0; i<sizeof(crtcparam[mode]); i++) {
		vid->crtca = i;
		vid->crtcd = softcrtc[i] = crtcparam[mode][i];
	}
	if(mode == USvision)
		CWR(ctl1,softvid.ctl1 & ~EurStandard);
	else
		CWR(ctl1,softvid.ctl1 | EurStandard);
}

static void
luts(int map, int colour, int ramp, int value)
{
	uchar *p;
	int i;

	p = vidbase + VIDREG + (colour*LUTSIZE) + (map*LUTMAPSIZE);
	for(i=0; i<LUTMAPSIZE; i++) {
		*p++ = value;
		value += ramp;
	}
}

static void
setquad(int q)
{
	Video *vid = vidregs;

	if(q != curquad) {
		curquad = q;
		CWR(ctl0, (softvid.ctl0 & ~Quadno) | ((q&3)<<4));
	}
}

static void
setcontin(int q)
{
	Video *vid = vidregs;

	setquad(q);
	CWR(ctl1, softvid.ctl1 | Continuous);
}

static void
setsnap(int q)
{
	Video *vid = vidregs;

	setquad(q);
	CWR(ctl1, softvid.ctl1 & ~Continuous);
}

static void
snap(int q)
{
	Video *vid = vidregs;

	setquad(q);
	if((vid->status & Grabbing) == 0)
		vid->snap = 0xff;
}

/*
 * copy data to/from frame-buffer's memory;
 * memmove can't be used because the device handles D16 but not D32.
 * the processor should switch to D8 cycles when the VME source/dest
 * is on an odd boundary.
 */
static void
vmecpy(void *a, void *b, long count)
{
	ushort *to, *from;
	int n, odd;

	if(count <= 0)
		return;
 	to = (ushort*)a;
	from = (ushort*)b;
	odd = count & 1;
	if((count >>= 1) != 0) {
		n = (count+7) >> 3;
		switch(count&7) {	/* Duff's device */
		case 0: do {	*to++ = *from++;
		case 7:		*to++ = *from++;
		case 6:		*to++ = *from++;
		case 5:		*to++ = *from++;
		case 4:		*to++ = *from++;
		case 3:		*to++ = *from++;
		case 2:		*to++ = *from++;
		case 1:		*to++ = *from++;
			} while(--n > 0);
		}
	}
	if(odd)
		*(uchar*)to = *(uchar*)from;
}
